/*
 * Segment 5 Intelligente Analyseverfahren, SWE 5.1 Ganglinienprognose
 * Copyright (C) 2007 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.glprog.glprog;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.fail;

import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import com.bitctrl.Constants;
import com.bitctrl.resource.Configuration;
import com.bitctrl.util.logging.LoggerTools;

import de.bsvrz.iav.gllib.gllib.Stuetzstelle;
import de.bsvrz.iav.gllib.gllib.dav.GanglinieMQ;
import de.bsvrz.iav.gllib.gllib.dav.GanglinienMQOperationen;
import de.bsvrz.iav.gllib.gllib.dav.Ganglinienprognose;
import de.bsvrz.iav.gllib.gllib.dav.GlProgAnfrage;
import de.bsvrz.iav.gllib.gllib.dav.GlProgAntwortEvent;
import de.bsvrz.iav.gllib.gllib.dav.GlProgAntwortListener;
import de.bsvrz.iav.gllib.gllib.dav.Messwerte;
import de.bsvrz.sys.funclib.bitctrl.modell.DatensendeException;

/**
 * Basisklasse für einen Testfall der Straßensubsegmentanalyse.
 *
 * @author BitCtrl Systems GmbH, Falko Schumann
 */
@SuppressWarnings("nls")
public abstract class AbstractTestfall {

	/**
	 * Empänger für die Istwerte.
	 */
	private class Empfaenger implements GlProgAntwortListener {

		/**
		 * Trägt die empfange Antwort in die Liste der Ist-Werte ein.
		 */
		@Override
		public void antwortEingetroffen(final GlProgAntwortEvent e) {
			if (e.getAbsenderZeichen().equals(getAbsenderZeichen())) {
				istWerte.addAll(e.getGanglinien());
				log.finer("Neue Prognoseantwort: " + e.getGanglinien());
			}
			if (istWerte.size() >= sollWerte.size()) {
				synchronized (this) {
					notify();
				}
			}

		}
	}

	/** Die Verbindung zur Datenbank. */
	private final Connection connection;

	/** Der Logger der Klasse. */
	private final Logger log;

	/** Der Empfänger der Ist-Werte. */
	private final Empfaenger empfaenger;

	/** Hält die Soll-Werte. */
	private final List<GanglinieMQ> sollWerte;

	/** Sichert die Ist-Werte bis zur Auswertung. */
	private final List<GanglinieMQ> istWerte;

	/**
	 * Initialisiert das Objekt.
	 *
	 * @throws SQLException
	 *             bei einem Datenbankfehler.
	 */
	public AbstractTestfall() throws SQLException {
		final Configuration konfig = Configuration.getConfiguration();
		final Level level = Level.parse(konfig.getString("debug.level"));
		LoggerTools.setLoggerLevel(level);
		log = Logger.getLogger(getClass().getName());
		log.setLevel(level);

		istWerte = new ArrayList<>();
		empfaenger = new Empfaenger();

		connection = initConnection();
		sollWerte = abfragenAntwortenSoll();
	}

	/**
	 * Initialisiert die Verbindung zur Datenbank.
	 *
	 * @return die offene Datenbankverbindung.
	 */
	private Connection initConnection() {
		Connection conn;

		log.fine("Lade Datenbanktreiber ...");
		try {
			Class.forName("org.hsqldb.jdbcDriver");
		} catch (final ClassNotFoundException ex) {
			throw new IllegalStateException("Datenbanktreiber nicht gefunden: " + ex);
		}

		log.fine("Verbinde mit Datenbank ...");
		try {
			conn = DriverManager.getConnection(
					"jdbc:hsqldb:res:de/bsvrz/iav/glprog/glprog/prueffall4;shutdown=true", "sa",
					"");
		} catch (final SQLException ex) {
			throw new IllegalStateException("Kann keine Verbindung zur Datenbank aufbauen: " + ex);
		}

		return conn;
	}

	/**
	 * Gibt die Liste der Sollwerte zurück.
	 *
	 * @return die Sollwerte.
	 */
	public List<GanglinieMQ> getSollWerte() {
		return sollWerte;
	}

	/**
	 * Gibt die Liste der Istwerte zurück.
	 *
	 * @return die Istwerte.
	 */
	public List<GanglinieMQ> getIstWerte() {
		return istWerte;
	}

	/**
	 * Führt den Test aus.
	 */
	public void ausfuehren() {
		Ganglinienprognose glProg;

		glProg = Ganglinienprognose.getInstanz();
		glProg.addAntwortListener(empfaenger);

		try {
			Thread.sleep(10 * Constants.MILLIS_PER_SECOND);
		} catch (final InterruptedException ex) {
			ex.printStackTrace();
		}

		log.info("Sende Ganglinienprognoseanfrage ...");
		try {
			glProg.sendeAnfrage(getClass().getName(), abfragenAnfragen());
		} catch (final DatensendeException ex) {
			ex.printStackTrace();
			fail("Konnte Testdaten nicht senden: " + ex);
		} catch (final SQLException ex) {
			ex.printStackTrace();
			fail("Konnte Testdaten nicht aus Datenbank abrufen: " + ex);
		}
	}

	/**
	 * Vergleicht die empfangen Ist-Werte mit den Soll-Werten.
	 */
	public void auswerten() {
		// Warten bis alle Antworten empfangen wurden.
		log.fine("Warte auf Eingang aller Ist-Werte ...");
		if (istWerte.size() < sollWerte.size()) {
			try {
				synchronized (empfaenger) {
					empfaenger.wait();
				}
			} catch (final InterruptedException ex) {
				ex.printStackTrace();
			}
		}

		log.fine("Vergleiche Soll- und Ist-Werte ...");

		assertEquals("Die Anzahl der empfangenen Ist-Werte stimmt nicht.", sollWerte.size(),
				istWerte.size());

		for (int i = 0; i < sollWerte.size(); ++i) {
			final GanglinieMQ sollGl, istGl;
			final List<Stuetzstelle<Messwerte>> soll, ist;

			log.finer("Soll-Ganglinie: "
					+ GanglinienMQOperationen.formatierterText(sollWerte.get(i)));
			log.finer(
					"Ist-Ganglinie: " + GanglinienMQOperationen.formatierterText(istWerte.get(i)));

			try {
				Thread.sleep(5000);
			} catch (final InterruptedException ex) {
				ex.printStackTrace();
			}

			assertEquals("Messquerschnitt der Ganglinie stimmt nicht",
					sollWerte.get(i).getMessQuerschnitt(), istWerte.get(i).getMessQuerschnitt());
			assertEquals("Approximation der Ganglinie stimmt nicht.",
					sollWerte.get(i).getApproximationDaK(), istWerte.get(i).getApproximationDaK());
			assertEquals("B-Spline-Ordnung stimmt nicht.", sollWerte.get(i).getBSplineOrdnung(),
					istWerte.get(i).getBSplineOrdnung());

			sollGl = sollWerte.get(i);
			istGl = istWerte.get(i);

			/*
			 * Ist- und Soll-Ganglinie auf 01.01.1970 normieren und dann die Stützstellen
			 * vergleichen.
			 */
			istGl.setPrognoseZeitraum(null);
			GanglinienMQOperationen.verschiebe(istGl,
					-istGl.getStuetzstellen().get(0).getZeitstempel());
			GanglinienMQOperationen.verschiebe(sollGl,
					-sollGl.getStuetzstellen().get(0).getZeitstempel());
			soll = sollGl.getStuetzstellen();
			ist = istGl.getStuetzstellen();
			assertEquals("Die Anzahl der Stützstellen der Ganglinie mit dem Index " + i
					+ " stimmt nicht.", soll.size(), ist.size());
			for (int j = 0; j < soll.size(); ++j) {
				assertEquals("Die Stützstelle mit dem Index " + j + " der Ganglinie mit dem Index "
						+ i + " ist falsch.", soll.get(j), ist.get(j));
			}
		}
	}

	/**
	 * Ruft die Sollwerte aus der Datenbank ab.
	 *
	 * @return die Sollwerte.
	 * @throws SQLException
	 *             bei einem Datenbankfehler.
	 */
	protected abstract List<GanglinieMQ> abfragenAntwortenSoll() throws SQLException;

	/**
	 * Ruft die Testdaten aus der Datenbank ab.
	 *
	 * @return die Testdaten.
	 * @throws SQLException
	 *             bei einem Datenbankfehler.
	 */
	protected abstract List<GlProgAnfrage> abfragenAnfragen() throws SQLException;

	/**
	 * Gibt das Absenderzeichen für das Senden der Ganglinienprognoseanfrage zurück. Die
	 * Standardimplementierung gibt den Klassennamen zurück.
	 *
	 * @return das Absenderzeichen.
	 */
	public String getAbsenderZeichen() {
		return getClass().getName();
	}

	/**
	 * Gibt die Verbindung zur Datenbank zurück.
	 *
	 * @return die Datenbankverbindung.
	 */
	public Connection getConnection() {
		return connection;
	}

	/**
	 * Gibt den Logger der Klasse zurück.
	 *
	 * @return der Logger.
	 */
	public Logger getLog() {
		return log;
	}

}
