/*
 * Segment 5 Intelligente Analyseverfahren, SWE 5.1 Ganglinienprognose
 * Copyright (C) 2007 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.glprog.glprog;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;

import java.sql.SQLException;
import java.util.Calendar;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.junit.AfterClass;
import org.junit.BeforeClass;
import org.junit.Test;

import com.bitctrl.Constants;
import com.bitctrl.resource.Configuration;
import com.bitctrl.util.Interval;
import com.bitctrl.util.logging.LoggerTools;

import de.bsvrz.iav.gllib.gllib.Stuetzstelle;
import de.bsvrz.iav.gllib.gllib.dav.GanglinieMQ;
import de.bsvrz.iav.gllib.gllib.dav.GanglinienMQOperationen;
import de.bsvrz.iav.gllib.gllib.dav.GlProgAnfrage;
import de.bsvrz.iav.gllib.gllib.dav.Messwerte;
import de.bsvrz.iav.gllib.gllib.junit.ArchivdatenFactory;
import de.bsvrz.iav.gllib.gllib.junit.EreignisFactory;
import de.bsvrz.iav.gllib.gllib.junit.GanglinienFactory;
import de.bsvrz.iav.gllib.gllib.modell.parameter.PdGanglinienModellPrognose;
import de.bsvrz.sys.funclib.bitctrl.kalender.Ereigniskalender;
import de.bsvrz.sys.funclib.bitctrl.modell.ObjektFactory;
import de.bsvrz.sys.funclib.bitctrl.modell.kalender.objekte.Ereignis;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.objekte.MessQuerschnittAllgemein;

/**
 * Führt den Prüffall 2 <em>Ganglinienprognose</em> aus. Es wird die zyklische Prognose getestet.
 * <p>
 * <em>Hinweis:</em> Es wird ein Messquerschnitt für den Test benötigt. Der Ereigniskalender und das
 * Archiv müssen laufen. Das Archiv muss so parametriert sein, dass es Kurzzeitdaten
 * (atg.verkehrsDatenKurzZeitMq) für den Testmessquerschnitt sichert, darf aber hierfür noch keine
 * Daten enthalten.
 *
 * @author BitCtrl Systems GmbH, Falko Schumann
 */
@SuppressWarnings("nls")
public class Prueffall2 extends AbstractTestfall {

	/** Der Logger der Klassse. */
	private static Logger log = Logger.getLogger(Prueffall2.class.getName());

	/** Die Factory für die Kalenderdaten des Tests. */
	private static EreignisFactory ereignisFactory;

	/** Die Factory für die Ganglinien des Tests. */
	private static GanglinienFactory ganglinienFactory;

	/** Der Messquerschnitt an dem der Test stattfindet. */
	private static MessQuerschnittAllgemein mq;

	/** Der Parameter der Ganglinienprognose am Messquerschnitt. */
	private static PdGanglinienModellPrognose param;

	/**
	 * Führt Vorbereitungen durch, die einmalig vor dem Testfall notwendig sind.
	 * <ul>
	 * <li>Der Messquerschnitt für den Test wird bestimmt (siehe Configuration.properties).
	 * <em>Bereits vorhandene Ganglinien an diesem
	 * Messquerschnitt werden durch den Test überschrieben!</em></li>
	 * <li>Die Testereignisse und -ereignistypen werden angelegt.</li>
	 * <li>Die für den Test am Messquerschnitt benötigten Ganglinien werden parametriert (@code
	 * atg.ganglinie).</li>
	 * <li>Prüft ob der Prognoseparameter am Messquerschnitt gesetzt ist und setzt auf die die
	 * notwendigen Daten.</li>
	 * <li>Schreibt die benötigten Archivdaten.</li>
	 * </ul>
	 *
	 * @throws Exception
	 *             bei einem Fehler.
	 */
	@BeforeClass
	public static void beforeClass() throws Exception {
		final Configuration konfig = Configuration.getConfiguration();
		final Level level = Level.parse(konfig.getString("debug.level"));
		LoggerTools.setLoggerLevel(level);
		log.setLevel(level);

		DavUtil.ensureConnected();
		final ObjektFactory factory = ObjektFactory.getInstanz();

		// Testmessquerschnitt bestimmen und parametrieren
		mq = (MessQuerschnittAllgemein) factory
				.getModellobjekt(konfig.getString("prueffall2.objekt.mq"));
		assertNotNull("Der Testmessquerschnitt existiert nicht.", mq);

		// Ereignisse und Ereignistypen anlegen
		ereignisFactory = new EreignisFactory(
				"jdbc:hsqldb:res:de/bsvrz/iav/glprog/glprog/prueffall2;shutdown=true", "sa", "");
		if (konfig.getBoolean("reset")) {
			ereignisFactory.aufraeumen();
		}
		ereignisFactory.anlegenEreignistypen();
		ereignisFactory.anlegenEreignisse();

		// Ganglinien anlegen
		ganglinienFactory = new GanglinienFactory(mq,
				"jdbc:hsqldb:res:de/bsvrz/iav/glprog/glprog/prueffall2;shutdown=true", "sa", "",
				konfig.getString("dav.db"));
		if (konfig.getBoolean("reset")) {
			ganglinienFactory.aufraeumen();
			factory.getVerbindung().sleep(Constants.MILLIS_PER_SECOND);
		}
		ganglinienFactory.anlegenGanglinien(GanglinienFactory.ERSTER_TAG);

		// Archivdaten anlegen
		if (konfig.getBoolean("archivdatenSchreiben")) {
			final ArchivdatenFactory archivdatenFactory;

			archivdatenFactory = new ArchivdatenFactory(mq,
					"jdbc:hsqldb:res:de/bsvrz/iav/glprog/glprog/prueffall2;shutdown=true", "sa",
					"");
			archivdatenFactory.anlegenArchivdaten();
		} else {
			log.info("Überspringe anlegen der Archivdaten.");
		}

		// Prognoseparameter setzen bzw. aktualisieren
		param = mq.getParameterDatensatz(PdGanglinienModellPrognose.class);
		param.anmeldenSender();
		PdGanglinienModellPrognose.Daten datum = param.abrufenDatum();
		datum = param.erzeugeDatum();
		datum.setAuswahlMethode(PdGanglinienModellPrognose.Daten.REFERENZGANGLINIE);
		datum.setMatchingIntervall(Constants.MILLIS_PER_HOUR);
		datum.setMaxMatchingFehler(100);
		datum.setPatternMatchingHorizont(6 * Constants.MILLIS_PER_HOUR);
		datum.setPatternMatchingOffset(6 * Constants.MILLIS_PER_HOUR);
		param.sendeDaten(datum);

		log.info("Warte 60 Sekunden ...");
		factory.getVerbindung().sleep(60 * Constants.MILLIS_PER_SECOND);
	}

	/**
	 * Räumt nach dem Test auf.
	 * <ul>
	 * <li>Die Testereignisse und -ereignistypen werden wieder gelöscht.</li>
	 * <li>Die Testganglinien am Messquerschnitt werden wieder entfernt.</li>
	 * </ul>
	 *
	 * @throws Exception
	 *             bei einem Fehler.
	 */
	@AfterClass
	public static void afterClass() throws Exception {
		ereignisFactory.aufraeumen();
		ganglinienFactory.aufraeumen();
	}

	/**
	 * Ruft nur den Superkonstruktor auf. Ist nur notwendig wegen der Exception die der
	 * Superkonstruktor werfen kann.
	 *
	 * @throws SQLException
	 *             bei einem Datenbankfehler.
	 */
	public Prueffall2() throws SQLException {
		super();
	}

	/**
	 * Führt eine zyklische Anfrage aus. Es werden drei Zyklen durchgeführt.
	 * <ol>
	 * <li>Im ersten Zyklus (Initialprognose) wird ein Pattern-Matching durchgeführt. Es liegen ein
	 * absolutes, ein additives und ein multiplikatives Ereignis an.</li>
	 * <li>Im zweiten Zyklus wurden die relativen Ereignisse entfernt. Es wird erneut ein
	 * Pattern-Matching durchgeführt.</li>
	 * <li>Im dritten Zyklus wurde der Prognoseparameter geändert. Der maximale Abstand für das
	 * Pattern-Matching wurde soweit gesenkt, dass keines mehr durchgeführt wird.</li>
	 * </ol>
	 *
	 * @throws Exception
	 *             bei einem Fehler.
	 */
	@Test(timeout = 10 * 60 * 1000)
	public void testPrognoseanfrage() throws Exception {
		final ObjektFactory factory = ObjektFactory.getInstanz();

		ausfuehren();

		// Ereignis löschen
		log.info("Warte 2 Minuten ...");
		factory.getVerbindung().sleep(2 * Constants.MILLIS_PER_MINUTE);
		Ereignis ereignis = (Ereignis) factory.getModellobjekt("ereignis.testMesseTag0");
		Ereigniskalender.getInstanz().loeschen(ereignis);
		ereignis = (Ereignis) factory.getModellobjekt("ereignis.testUnfallTag0");
		Ereigniskalender.getInstanz().loeschen(ereignis);
		log.fine("Relatives Ereignis entfernt.");

		// Parameter ändern
		log.info("Warte 30 Sekunden ...");
		factory.getVerbindung().sleep(30 * Constants.MILLIS_PER_SECOND);
		final PdGanglinienModellPrognose.Daten datum = param.abrufenDatum();
		datum.setMaxMatchingFehler(5);
		param.sendeDaten(datum);
		log.fine("Prognoseparameter geändert.");

		auswerten();
	}

	/**
	 * Berechnet einen Punkt der Ganglinie. Der Punkt muss mit dem vorher berechneten Werten
	 * übereinstimmen.
	 */
	@Test
	public void testPunktDerGanglinieAusrechnen() {
		final GanglinieMQ ganglinie;
		final Stuetzstelle<Messwerte> soll, ist;
		final long zeitstempel;

		log.info("Berechne einen Punkt der Ganglinie ...");

		zeitstempel = (20 * Constants.MILLIS_PER_HOUR) + (30 * Constants.MILLIS_PER_MINUTE);
		soll = new Stuetzstelle<>(zeitstempel,
				new Messwerte(2226.177, 1478.627, 190.872, 126.159));

		ganglinie = getSollWerte().get(0);
		ist = ganglinie.getStuetzstelle(zeitstempel);

		assertEquals(soll.getZeitstempel(), ist.getZeitstempel());
		assertEquals(soll.getWert().getQKfz(), ist.getWert().getQKfz(), 0.001);
		assertEquals(soll.getWert().getQLkw(), ist.getWert().getQLkw(), 0.001);
		assertEquals(soll.getWert().getVPkw(), ist.getWert().getVPkw(), 0.001);
		assertEquals(soll.getWert().getVLkw(), ist.getWert().getVLkw(), 0.001);
	}

	/**
	 * Versucht eine Ganglinie mit Polylinieapproximation auf die Prognoseganglinie zu addieren.
	 */
	@Test(expected = IllegalArgumentException.class)
	public void testAdditionPolylinie() {
		final GanglinieMQ prognoseGl, gl;

		log.info("Addiere eine andere Ganglinie mit Polylinieapproximation ...");

		prognoseGl = getSollWerte().get(0);
		gl = getSollWerte().get(1);
		gl.setApproximationDaK(GanglinieMQ.APPROX_POLYLINE);
		GanglinienMQOperationen.addiere(prognoseGl, gl);
	}

	/**
	 * Versucht eine Ganglinie mit Polylinieapproximation mit der Prognoseganglinie zu
	 * multiplizieren.
	 */
	@Test(expected = IllegalArgumentException.class)
	public void testMultiplikationPolylinie() {
		final GanglinieMQ prognoseGl, gl;

		log.info("Addiere eine andere Ganglinie mit Polylinieapproximation ...");

		prognoseGl = getSollWerte().get(0);
		gl = getSollWerte().get(1);
		gl.setApproximationDaK(GanglinieMQ.APPROX_POLYLINE);
		GanglinienMQOperationen.multipliziere(prognoseGl, gl);
	}

	@Override
	protected List<GlProgAnfrage> abfragenAnfragen() throws SQLException {
		GlProgAnfrage anfrage;
		long start, ende;
		Calendar calendar;

		calendar = Calendar.getInstance();
		calendar.set(Calendar.HOUR_OF_DAY, 12);
		calendar.set(Calendar.MINUTE, 0);
		calendar.set(Calendar.SECOND, 0);
		calendar.set(Calendar.MILLISECOND, 0);
		start = calendar.getTimeInMillis();
		calendar.add(Calendar.HOUR_OF_DAY, 12);
		ende = calendar.getTimeInMillis();

		anfrage = new GlProgAnfrage(mq, new Interval(start, ende), false, true,
				1 * Constants.MILLIS_PER_MINUTE, 1, 3 * Constants.MILLIS_PER_MINUTE);

		return Collections.singletonList(anfrage);
	}

	@Override
	protected List<GanglinieMQ> abfragenAntwortenSoll() throws SQLException {
		List<GanglinieMQ> soll;
		Iterator<GanglinieMQ> iterator;

		soll = ganglinienFactory.getGanglinienTestdaten(1);
		iterator = soll.iterator();
		while (iterator.hasNext()) {
			if (iterator.next().size() == 0) {
				iterator.remove();
			}
		}

		assertEquals("Die Anzahl der Prognoseganglinie ist falsch.", 3, soll.size());
		return soll;
	}

}
