/*
 * Segment 5 Intelligente Analyseverfahren, SWE 5.1 Ganglinienprognose
 * Copyright (C) 2007 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.glprog.glprog;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;

import java.sql.SQLException;
import java.util.Calendar;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.junit.AfterClass;
import org.junit.BeforeClass;
import org.junit.Test;

import com.bitctrl.Constants;
import com.bitctrl.resource.Configuration;
import com.bitctrl.util.Interval;
import com.bitctrl.util.Timestamp;

import de.bsvrz.iav.gllib.gllib.dav.GanglinieMQ;
import de.bsvrz.iav.gllib.gllib.dav.GanglinienMQOperationen;
import de.bsvrz.iav.gllib.gllib.dav.GlProgAnfrage;
import de.bsvrz.iav.gllib.gllib.junit.EreignisFactory;
import de.bsvrz.iav.gllib.gllib.junit.GanglinienFactory;
import de.bsvrz.iav.gllib.gllib.modell.parameter.PdGanglinienModellPrognose;
import de.bsvrz.sys.funclib.bitctrl.modell.ObjektFactory;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.objekte.MessQuerschnittAllgemein;

/**
 * Führt den Prüffall 3 <em>Ganglinienprognose</em> aus.
 * <p>
 * Es wird die Ganglinienprognose an einem Ostermontag durchgeführt. Folgende Ereignisse liegen an:
 *
 * <table summary="Liste der vordefinierten Ereignisse">
 * <tr>
 * <th>Ereignis</th>
 * <th>Priorität</th>
 * <th>historische Ganglinien</th>
 * </tr>
 * <tr>
 * <td>Ostermontag</td>
 * <td>100</td>
 * <td>keine vorhanden</td>
 * </tr>
 * <tr>
 * <td>Montagsfeiertag</td>
 * <td>80</td>
 * <td>mehrere vorhanden</td>
 * </tr>
 * <tr>
 * <td>Feiertag</td>
 * <td>60</td>
 * <td>keine vorhanden</td>
 * </tr>
 * <tr>
 * <td>Montag</td>
 * <td>20</td>
 * <td>mehrere vorhanden</td>
 * </tr>
 * </table>
 * <p>
 * Die historischen Ganglinien sind alle B-Splines höchstmöglicher Ordnung. Die Ordnung ist hier
 * laut Prüfspezifikation auf 50 festgelegt.
 * <p>
 * Für den Test wird ein Messquerschnitt benötigt. Das Archiv wird nicht verwendet.
 *
 * @author BitCtrl Systems GmbH, Falko Schumann
 */
@SuppressWarnings("nls")
public class Prueffall3 extends AbstractTestfall {

	/** Der Logger der Klassse. */
	private static Logger log = Logger.getLogger(Prueffall3.class.getName());

	/** Die Factory für die Kalenderdaten des Tests. */
	private static EreignisFactory ereignisFactory;

	/** Die Factory für die Ganglinien des Tests. */
	private static GanglinienFactory ganglinienFactory;

	/** Der Messquerschnitt an dem der Test stattfindet. */
	private static MessQuerschnittAllgemein mq;

	/**
	 * Führt Vorbereitungen durch, die einmalig vor dem Testfall notwendig sind.
	 * <ul>
	 * <li>Der Messquerschnitt für den Test wird bestimmt (siehe Configuration.properties).
	 * <em>Bereits vorhandene Ganglinien an diesem
	 * Messquerschnitt werden durch den Test überschrieben!</em></li>
	 * <li>Die Testereignisse und -ereignistypen werden angelegt.</li>
	 * <li>Die für den Test am Messquerschnitt benötigten Ganglinien werden parametriert (@code
	 * atg.ganglinie).</li>
	 * <li>Prüft ob der Prognoseparameter am Messquerschnitt gesetzt ist und ob als Auswahlkriterium
	 * "Referenzganglinie" angegeben ist. Ist der Parameter bereits vorhanden, wird lediglich die
	 * Auswahlmethode auf Referenzganglinie geändert.</li>
	 * </ul>
	 *
	 * @throws Exception
	 *             bei einem Fehler.
	 */
	@BeforeClass
	public static void beforeClass() throws Exception {
		final Configuration konfig;
		final Level level;
		final PdGanglinienModellPrognose param;
		PdGanglinienModellPrognose.Daten datum;

		konfig = Configuration.getConfiguration();
		level = Level.parse(konfig.getString("debug.level"));
		log.setLevel(level);

		DavUtil.ensureConnected();
		final ObjektFactory factory = ObjektFactory.getInstanz();

		// Testmessquerschnitt bestimmen und parametrieren
		mq = (MessQuerschnittAllgemein) factory
				.getModellobjekt(konfig.getString("prueffall3.objekt.mq"));
		assertNotNull("Der Testmessquerschnitt existiert nicht.", mq);

		// Ereignisse und Ereignistypen anlegen
		ereignisFactory = new EreignisFactory(
				"jdbc:hsqldb:res:de/bsvrz/iav/glprog/glprog/prueffall3;shutdown=true", "sa", "");
		if (konfig.getBoolean("reset")) {
			ereignisFactory.aufraeumen();
		}
		ereignisFactory.anlegenEreignistypen();
		ereignisFactory.anlegenEreignisse();

		// Ganglinien anlegen
		ganglinienFactory = new GanglinienFactory(mq,
				"jdbc:hsqldb:res:de/bsvrz/iav/glprog/glprog/prueffall3;shutdown=true", "sa", "",
				konfig.getString("dav.db"));
		if (konfig.getBoolean("reset")) {
			ganglinienFactory.aufraeumen();
			factory.getVerbindung().sleep(Constants.MILLIS_PER_SECOND);
		}
		ganglinienFactory.anlegenGanglinien(GanglinienFactory.ERSTER_TAG);

		// Prognoseparameter setzen
		param = mq.getParameterDatensatz(PdGanglinienModellPrognose.class);
		param.anmeldenSender();
		datum = param.abrufenDatum();
		if (datum.isValid()) {
			if (datum.getAuswahlMethode() != PdGanglinienModellPrognose.Daten.REFERENZGANGLINIE) {
				log.warning("Der Parameter existiert, muss aber angepasst werden.");
				datum.setAuswahlMethode(PdGanglinienModellPrognose.Daten.REFERENZGANGLINIE);
			}
		} else {
			datum = param.erzeugeDatum();
			datum.setAuswahlMethode(PdGanglinienModellPrognose.Daten.REFERENZGANGLINIE);
		}
		param.sendeDaten(datum);

		log.info("Warte 60 Sekunden ...");
		factory.getVerbindung().sleep(60 * Constants.MILLIS_PER_SECOND);
	}

	/**
	 * Räumt nach dem Test auf.
	 * <ul>
	 * <li>Die Testereignisse und -ereignistypen werden wieder gelöscht.</li>
	 * <li>Die Testganglinien am Messquerschnitt werden wieder entfernt.</li>
	 * </ul>
	 *
	 * @throws Exception
	 *             bei einem Fehler.
	 */
	@AfterClass
	public static void afterClass() throws Exception {
		ereignisFactory.aufraeumen();
		ganglinienFactory.aufraeumen();
	}

	/**
	 * Ruft nur den Superkonstruktor auf. Ist nur notwendig wegen der Exception die der
	 * Superkonstruktor werfen kann.
	 *
	 * @throws SQLException
	 *             bei einem Datenbankfehler.
	 */
	public Prueffall3() throws SQLException {
		super();
	}

	/**
	 * Führt eine einmalige Anfrage für eine langfristige Prognose an die Ganglinienprognose aus. Im
	 * Ergebnis werden nur die Verfahren der langfristigen Prognose verwendet. Die als
	 * Referenzganglinie ausgezeichnten Ganglinien werden für die Prognose verwendet.
	 *
	 * @throws Exception
	 *             bei einem Fehler.
	 */
	@Test(timeout = 60 * 1000)
	public void testPrognoseanfrage() throws Exception {
		ausfuehren();
		auswerten();
	}

	/**
	 * Addiert eine Ganglinie auf die Prognoseganglinie. Beide Ganglinien überschneiden sich nur
	 * teilweise. Die Zeitstempel der Stützstellen im Bereich der Überschneidung sind bis auf eine
	 * identisch.
	 * <p>
	 * Berechnet den Basisabstand zwischen der Prognoseganglinie und der verknüpften.
	 *
	 * @throws SQLException
	 *             bei einem Fehler.
	 */
	@Test
	public void testAddition() throws SQLException {
		final GanglinieMQ prognoseGl, summand, soll, ist;

		prognoseGl = abfragenAntwortenSoll().get(0);
		summand = ganglinienFactory.getGanglinienTestdaten(2).get(0);
		soll = ganglinienFactory.getGanglinienTestdaten(3).get(0);

		ist = GanglinienMQOperationen.addiere(prognoseGl, summand);

		assertEquals("Anzahl der Stützstellen stimmt nicht.", soll.size(), ist.size());
		for (final long t : soll.keySet()) {
			assertEquals("Der Wert für " + Timestamp.absoluteTime(t) + " ist falsch.", soll.get(t),
					ist.get(t));
		}

		assertEquals("Basisabstand ist nicht korrekt.", 66,
				GanglinienMQOperationen.basisabstand(prognoseGl, soll));
	}

	/**
	 * Subtrahiert eine Ganglinie von der Prognoseganglinie. Beide Ganglinien überschneiden sich nur
	 * teilweise. Die Zeitstempel der Stützstellen im Bereich der Überschneidung sind bis auf eine
	 * identisch.
	 * <p>
	 * Berechnet den Basisabstand zwischen der Prognoseganglinie und der verknüpften.
	 *
	 * @throws SQLException
	 *             bei einem Fehler.
	 */
	@Test
	public void testSubtraktion() throws SQLException {
		final GanglinieMQ prognoseGl, subtrahend, soll, ist;

		prognoseGl = abfragenAntwortenSoll().get(0);
		subtrahend = ganglinienFactory.getGanglinienTestdaten(2).get(0);
		soll = ganglinienFactory.getGanglinienTestdaten(3).get(1);

		ist = GanglinienMQOperationen.subtrahiere(prognoseGl, subtrahend);

		assertEquals("Anzahl der Stützstellen stimmt nicht.", soll.size(), ist.size());
		for (final long t : soll.keySet()) {
			assertEquals("Der Wert für " + Timestamp.absoluteTime(t) + " ist falsch.", soll.get(t),
					ist.get(t));
		}

		assertEquals("Basisabstand ist nicht korrekt.", 177,
				GanglinienMQOperationen.basisabstand(prognoseGl, soll));
	}

	/**
	 * Multipliziert eine Ganglinie mit der Prognoseganglinie. Beide Ganglinien überschneiden sich
	 * nur teilweise. Die Zeitstempel der Stützstellen im Bereich der Überschneidung sind bis auf
	 * eine identisch.
	 * <p>
	 * Berechnet den Basisabstand zwischen der Prognoseganglinie und der verknüpften.
	 *
	 * @throws SQLException
	 *             bei einem Fehler.
	 */
	@Test
	public void testMultiplikation() throws SQLException {
		final GanglinieMQ prognoseGl, faktor, soll, ist;

		prognoseGl = abfragenAntwortenSoll().get(0);
		faktor = ganglinienFactory.getGanglinienTestdaten(2).get(0);
		soll = ganglinienFactory.getGanglinienTestdaten(3).get(2);

		ist = GanglinienMQOperationen.multipliziere(prognoseGl, faktor);

		assertEquals("Anzahl der Stützstellen stimmt nicht.", soll.size(), ist.size());
		for (final long t : soll.keySet()) {
			assertEquals("Der Wert für " + Timestamp.absoluteTime(t) + " ist falsch.", soll.get(t),
					ist.get(t));
		}

		assertEquals("Basisabstand ist nicht korrekt.", 197,
				GanglinienMQOperationen.basisabstand(prognoseGl, soll));
	}

	@Override
	protected List<GlProgAnfrage> abfragenAnfragen() throws SQLException {
		GlProgAnfrage anfrage;
		long start, ende;
		Calendar calendar;

		calendar = Calendar.getInstance();
		calendar.set(Calendar.HOUR_OF_DAY, 0);
		calendar.set(Calendar.MINUTE, 0);
		calendar.set(Calendar.SECOND, 0);
		calendar.set(Calendar.MILLISECOND, 0);
		start = calendar.getTimeInMillis();
		calendar.add(Calendar.HOUR_OF_DAY, 24);
		ende = calendar.getTimeInMillis();

		anfrage = new GlProgAnfrage(mq, new Interval(start, ende), true);

		return Collections.singletonList(anfrage);
	}

	@Override
	protected List<GanglinieMQ> abfragenAntwortenSoll() throws SQLException {
		final List<GanglinieMQ> soll;
		final Iterator<GanglinieMQ> iterator;

		soll = ganglinienFactory.getGanglinienTestdaten(1);
		iterator = soll.iterator();
		while (iterator.hasNext()) {
			if (iterator.next().size() == 0) {
				iterator.remove();
			}
		}

		assertEquals("Die Anzahl der Prognoseganglinie ist falsch.", 1, soll.size());
		return soll;
	}

}
