/*
 * Segment 5 Intelligente Analyseverfahren, SWE 5.1 Ganglinienprognose
 * Copyright (C) 2007 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.glprog.glprog;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;

import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.junit.AfterClass;
import org.junit.BeforeClass;
import org.junit.Test;

import com.bitctrl.Constants;
import com.bitctrl.resource.Configuration;
import com.bitctrl.util.Interval;

import de.bsvrz.iav.gllib.gllib.Stuetzstelle;
import de.bsvrz.iav.gllib.gllib.dav.GanglinieMQ;
import de.bsvrz.iav.gllib.gllib.dav.GlProgAnfrage;
import de.bsvrz.iav.gllib.gllib.dav.Messwerte;
import de.bsvrz.iav.gllib.gllib.junit.ArchivdatenFactory;
import de.bsvrz.iav.gllib.gllib.junit.EreignisFactory;
import de.bsvrz.iav.gllib.gllib.junit.GanglinienFactory;
import de.bsvrz.iav.gllib.gllib.modell.parameter.PdGanglinienModellPrognose;
import de.bsvrz.sys.funclib.bitctrl.modell.ObjektFactory;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.objekte.MessQuerschnittAllgemein;

/**
 * Führt den Prüffall 5 <em>Ganglinienprognose</em> aus.
 * <p>
 * Es werden zwei Prognosen durchgeführt. Für den gestrigen und heutigen Tag "TestMontag" existieren
 * drei relative Ganglinien (Polyline) für "TestMesse". Es liegen Archivdaten zum Abfragezeitpunkt
 * vor.
 * <p>
 * Die Anfrage wird zweimal durchgeführt:
 * <ol>
 * <li>Die prognostizierte Ganglinie ist identisch mit einer historischen Ganglinie.</li>
 * <li>Die prognostizierte Ganglinie ist identisch mit einer historischen Ganglinie, die aber
 * verschoben wurde.</li>
 * </ol>
 * <p>
 * Es wird ein Punkt der Prognoseganglinie berechnet.
 * <p>
 * Für den Test wird ein Messquerschnitt benötigt. Das Archiv wird verwendet und darf keine Daten
 * für den verwendeten Testmessquerschnitt enthalten.
 *
 *
 * @author BitCtrl Systems GmbH, Falko Schumann
 */
@SuppressWarnings("nls")
public class Prueffall5 extends AbstractTestfall {

	/** Der Logger der Klassse. */
	private static Logger log = Logger.getLogger(Prueffall5.class.getName());

	/** Die Factory für die Kalenderdaten des Tests. */
	private static EreignisFactory ereignisFactory;

	/** Die Factory für die Ganglinien des Tests. */
	private static GanglinienFactory ganglinienFactory;

	/** Der Messquerschnitt an dem der Test stattfindet. */
	private static MessQuerschnittAllgemein mq;

	private static String dbVerzeichnis;

	/**
	 * Führt Vorbereitungen durch, die einmalig vor dem Testfall notwendig sind.
	 * <ul>
	 * <li>Der Messquerschnitt für den Test wird bestimmt (siehe Configuration.properties).
	 * <em>Bereits vorhandene Ganglinien an diesem
	 * Messquerschnitt werden durch den Test überschrieben!</em></li>
	 * <li>Die Testereignisse und -ereignistypen werden angelegt.</li>
	 * <li>Die für den Test am Messquerschnitt benötigten Ganglinien werden parametriert (@code
	 * atg.ganglinie).</li>
	 * <li>Setzt den Prognoseparameter am Messquerschnitt.</li>
	 * </ul>
	 *
	 * @throws Exception
	 *             bei einem Fehler.
	 */
	@BeforeClass
	public static void beforeClass() throws Exception {
		final Configuration konfig;
		final Level level;
		final PdGanglinienModellPrognose param;
		final PdGanglinienModellPrognose.Daten datum;

		konfig = Configuration.getConfiguration();
		level = Level.parse(konfig.getString("debug.level"));
		log.setLevel(level);

		dbVerzeichnis = konfig.getString("dav.db");

		DavUtil.ensureConnected();
		final ObjektFactory factory = ObjektFactory.getInstanz();

		// Testmessquerschnitt bestimmen
		mq = (MessQuerschnittAllgemein) factory
				.getModellobjekt(konfig.getString("prueffall5.objekt.mq"));
		assertNotNull("Der Testmessquerschnitt existiert nicht.", mq);

		// Ereignisse und Ereignistypen anlegen
		ereignisFactory = new EreignisFactory(
				"jdbc:hsqldb:res:de/bsvrz/iav/glprog/glprog/prueffall5;shutdown=true", "sa", "");
		if (konfig.getBoolean("reset")) {
			ereignisFactory.aufraeumen();
		}
		ereignisFactory.anlegenEreignistypen();
		ereignisFactory.anlegenEreignisse();

		// Ganglinien anlegen
		ganglinienFactory = new GanglinienFactory(mq,
				"jdbc:hsqldb:res:de/bsvrz/iav/glprog/glprog/prueffall5;shutdown=true", "sa", "",
				konfig.getString("dav.db"));
		if (konfig.getBoolean("reset")) {
			ganglinienFactory.aufraeumen();
			factory.getVerbindung().sleep(Constants.MILLIS_PER_SECOND);
		}
		ganglinienFactory.anlegenGanglinien(GanglinienFactory.ERSTER_TAG);

		// Archivdaten anlegen
		if (konfig.getBoolean("archivdatenSchreiben")) {
			final ArchivdatenFactory archivdatenFactory;

			archivdatenFactory = new ArchivdatenFactory(mq,
					"jdbc:hsqldb:res:de/bsvrz/iav/glprog/glprog/prueffall5;shutdown=true", "sa",
					"");
			archivdatenFactory.anlegenArchivdaten();
		} else {
			log.info("Überspringe anlegen der Archivdaten.");
		}

		// Prognoseparameter setzen bzw. aktualisieren
		param = mq.getParameterDatensatz(PdGanglinienModellPrognose.class);
		param.anmeldenSender();
		datum = param.erzeugeDatum();
		datum.setAuswahlMethode(PdGanglinienModellPrognose.Daten.WAHRSCHEINLICHSTE_GANGLINIE);
		datum.setPatternMatchingHorizont(2 * Constants.MILLIS_PER_HOUR);
		datum.setPatternMatchingOffset(2 * Constants.MILLIS_PER_HOUR);
		datum.setMatchingIntervall(Constants.MILLIS_PER_MINUTE);
		datum.setMaxMatchingFehler(100);
		param.sendeDaten(datum);
		log.info("Prognoseparameter gesetzt.");

		log.info("Warte 60 Sekunden ...");
		factory.getVerbindung().sleep(60 * Constants.MILLIS_PER_SECOND);
	}

	/**
	 * Räumt nach dem Test auf.
	 * <ul>
	 * <li>Die Testereignisse und -ereignistypen werden wieder gelöscht.</li>
	 * <li>Die Testganglinien am Messquerschnitt werden wieder entfernt.</li>
	 * </ul>
	 *
	 * @throws Exception
	 *             bei einem Fehler.
	 */
	@AfterClass
	public static void afterClass() throws Exception {
		ereignisFactory.aufraeumen();
		ganglinienFactory.aufraeumen();
	}

	/**
	 * Ruft nur den Superkonstruktor auf. Ist nur notwendig wegen der Exception die der
	 * Superkonstruktor werfen kann.
	 *
	 * @throws SQLException
	 *             bei einem Datenbankfehler.
	 */
	public Prueffall5() throws SQLException {
		super();
	}

	/**
	 * Führt zwei Anfragen an die Ganglinienprognose aus. Es liegen drei verschiedene historische
	 * Ganglinien vor:
	 * <ul>
	 * <li>eine mit einfachen Messwerten</li>
	 * <li>eine mit den doppelten Messwerten</li>
	 * <li>eine einfachen Werten und einen Beule in der Mitte</li>
	 * </ul>
	 * Im Archiv ist ebenfalls ein Verlauf mit einfachen Messwerten und einer Beule in der Mitte.
	 * <p>
	 * Die erste Anfrage ist so getimt, dass in der Antwort die Beule aus der dritten Ganglinie und
	 * den Archivdaten übereinstimmen. Die zweite Anfrage ist so getimt, dass die dritte Ganglinie
	 * in der Antwort verschoben wird. Beides lässt sich anhand der Beule in der Prognoseganglinie
	 * nachvollziehen. In der ersten Prognoseganglinie ist die Beule an der selben Stelle wie in der
	 * historischen Ganglinie und in der zweiten Prognose ist diese Beule versetzt.
	 */
	@Test(timeout = 60 * 1000)
	public void testPatternMatching() {
		ausfuehren();
		auswerten();
	}

	/**
	 * Berechnet einen Punkt der Ganglinie. Der Punkt muss mit dem vorher berechneten Werten
	 * übereinstimmen.
	 */
	@Test
	public void testPunktDerGanglinieAusrechnen() {
		final GanglinieMQ ganglinie;
		final Stuetzstelle<Messwerte> soll, ist;
		final long zeitstempel;

		log.info("Berechne einen Punkt der Ganglinie ...");

		zeitstempel = (20 * Constants.MILLIS_PER_HOUR) + (30 * Constants.MILLIS_PER_MINUTE);
		soll = new Stuetzstelle<>(zeitstempel, new Messwerte(1500.0, 375.0, 150.0, 60.0));

		ganglinie = getSollWerte().get(0);
		ist = ganglinie.getStuetzstelle(zeitstempel);

		assertEquals(soll, ist);
	}

	@Override
	protected List<GlProgAnfrage> abfragenAnfragen() throws SQLException {
		final List<GlProgAnfrage> anfragen;
		final Calendar cal;
		final long start1, ende1, start2, ende2;

		// Gestern 0:00 Uhr
		cal = Calendar.getInstance();
		cal.set(Calendar.HOUR_OF_DAY, 0);
		cal.set(Calendar.MINUTE, 0);
		cal.set(Calendar.SECOND, 0);
		cal.set(Calendar.MILLISECOND, 0);
		cal.add(Calendar.DAY_OF_YEAR, -1);

		// Gestern 16:00 Uhr
		cal.add(Calendar.HOUR_OF_DAY, 16);
		start1 = cal.getTimeInMillis();

		// Gestern 24:00 Uhr bzw. heute 0:00 Uhr
		cal.add(Calendar.HOUR_OF_DAY, 8);
		ende1 = cal.getTimeInMillis();

		// Heute 17:00 Uhr
		cal.add(Calendar.HOUR_OF_DAY, 17);
		start2 = cal.getTimeInMillis();

		// Heute 24:00 Uhr
		cal.add(Calendar.HOUR_OF_DAY, 7);
		ende2 = cal.getTimeInMillis();

		anfragen = new ArrayList<>();
		anfragen.add(new GlProgAnfrage(mq, new Interval(start1, ende1), false));
		anfragen.add(new GlProgAnfrage(mq, new Interval(start2, ende2), false));

		return anfragen;
	}

	@Override
	protected List<GanglinieMQ> abfragenAntwortenSoll() throws SQLException {
		ganglinienFactory = new GanglinienFactory(mq,
				"jdbc:hsqldb:res:de/bsvrz/iav/glprog/glprog/prueffall5;shutdown=true", "sa", "",
				dbVerzeichnis);
		return ganglinienFactory.getGanglinienTestdaten(1);
	}

}
