/*
 * Segment 5 Intelligente Analyseverfahren, SWE 5.1 Ganglinienprognose
 * Copyright (C) 2007 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.glprog.glprog.util;

import java.util.Calendar;
import java.util.Collections;
import java.util.List;

import com.bitctrl.Constants;
import com.bitctrl.util.Interval;

import de.bsvrz.iav.gllib.gllib.dav.GlProgAnfrage;
import de.bsvrz.iav.glprog.glprog.GlProgMsg;
import de.bsvrz.sys.funclib.bitctrl.daf.BetriebsmeldungDaten;
import de.bsvrz.sys.funclib.bitctrl.daf.LogTools;
import de.bsvrz.sys.funclib.bitctrl.kalender.Ereigniskalender;
import de.bsvrz.sys.funclib.bitctrl.kalender.KalenderAnfrage;
import de.bsvrz.sys.funclib.bitctrl.kalender.KalenderEvent;
import de.bsvrz.sys.funclib.bitctrl.kalender.KalenderListener;
import de.bsvrz.sys.funclib.bitctrl.modell.DatensendeException;
import de.bsvrz.sys.funclib.bitctrl.modell.kalender.onlinedaten.OdEreignisKalenderAntwort.Daten.Zustandswechsel;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.objekte.NetzBestandTeil;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.objekte.StrassenSegment;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Hilfsklasse zum Bestimmen der relevanten Ereignisse zu einer Ganglinienanfrage.
 *
 * @author BitCtrl Systems GmbH, Falko Schumann
 */
public class EreignisHelper {

	private final Debug log = Debug.getLogger();

	private final KalenderListener kalenderListener = new KalenderListener() {

		@Override
		public void antwortEingetroffen(final KalenderEvent e) {
			if (e.getAbsenderZeichen().equals(absenderZeichen)) {
				LogTools.log(log, GlProgMsg.FinerKalenderAntwort, anfrage);

				e.getKalender().removeKalenderListener(this);
				kalenderantwort = e.getZustandswechsel();
				synchronized (EreignisHelper.this) {
					EreignisHelper.this.notify();
				}
			}
		}

	};

	private GlProgAnfrage anfrage;
	private String absenderZeichen;
	private List<Zustandswechsel> kalenderantwort;

	/**
	 * Gibt die Anfrage zurück, zu der die relevanten Ereignisse bestimmt werden sollen.
	 *
	 * @return die Anfrage.
	 */
	public GlProgAnfrage getAnfrage() {
		return anfrage;
	}

	/**
	 * Legt die Anfrage fest, zu der die relevanten Ereignisse bestimmt werden sollen.
	 *
	 * @param anfrage
	 *            die Anfrage.
	 */
	public void setAnfrage(final GlProgAnfrage anfrage) {
		this.anfrage = anfrage;
	}

	/**
	 * Gibt das Zeichen des Absenders der Prognoseanfrage zurück.
	 *
	 * @return das Absenderzeichen.
	 */
	public String getAbsenderZeichen() {
		return absenderZeichen;
	}

	/**
	 * Legt das Zeichen des Absenders der Prognoseanfrage fest.
	 *
	 * @param absenderZeichen
	 *            das Absenderzeichen.
	 */
	public void setAbsenderZeichen(final String absenderZeichen) {
		this.absenderZeichen = absenderZeichen;
	}

	/**
	 * Fragt nach den im Prognosezeitraum anliegenden Ereignissen. Die Anfrage an den Kalender wird
	 * so formuliert, dass sie 0:00 Uhr beginnt, damit die Kalenderantwort das Tagesereignis
	 * (z.&nbsp;B. Montag) nicht nur als Ende, sondern auch als Anfang enthält.
	 *
	 * @return {@code true}, wenn die Kalenderanfrage erfolgreich abgesetzt wurde, sonst
	 *         {@code false}.
	 */
	public List<Zustandswechsel> getEreignisse() {
		assert anfrage != null : "Anfrage ist null."; //$NON-NLS-1$
		assert absenderZeichen != null : "Absenderzeichen ist null."; //$NON-NLS-1$

		LogTools.log(log, GlProgMsg.FinerStarteKalenderAnfrage, anfrage);

		final Ereigniskalender kalender;
		final KalenderAnfrage kalenderAnfrage;
		final StrassenSegment ss;
		final Calendar cal;
		long start;

		// Anfrage an Kalender beginnt immer mit Tagesbeginn 0:00:00 Uhr
		start = anfrage.getPrognoseZeitraum().getStart();
		cal = Calendar.getInstance();
		cal.setTimeInMillis(start);
		cal.set(Calendar.HOUR_OF_DAY, 0);
		cal.set(Calendar.MINUTE, 0);
		cal.set(Calendar.SECOND, 0);
		cal.set(Calendar.MILLISECOND, 0);
		start = cal.getTimeInMillis();

		kalender = Ereigniskalender.getInstanz();
		if (!kalender.isBereit()) {
			LogTools.log(log, GlProgMsg.ErrorKalenderNichtVerfuegbar);
			return Collections.emptyList();
		}

		kalender.addKalenderListener(kalenderListener);
		kalenderAnfrage = new KalenderAnfrage();
		kalenderAnfrage.setIntervall(new Interval(start, anfrage.getPrognoseZeitraum().getEnd()));
		ss = anfrage.getMessQuerschnitt().getStrassenSegment();
		if (ss instanceof NetzBestandTeil) {
			kalenderAnfrage.getRaeumlicheGueltigkeit().add((NetzBestandTeil) ss);
		}
		try {
			kalender.sendeAnfrage(getAbsenderZeichen(), kalenderAnfrage);
		} catch (final DatensendeException ex) {
			LogTools.log(log, GlProgMsg.ErrorBeiKalenderanfrageSenden, ex.getLocalizedMessage());

			kalender.removeKalenderListener(kalenderListener);
			return Collections.emptyList();
		}

		// Maximal 60 Sekunden auf Ereigniskalender warten
		LogTools.log(log, GlProgMsg.FineWarteAufKalender, anfrage);
		synchronized (this) {
			try {
				wait(Constants.MILLIS_PER_MINUTE);
			} catch (final InterruptedException ex) {
				// kann ignoriert werden
			}
		}

		kalender.removeKalenderListener(kalenderListener);
		if (kalenderantwort == null) {
			LogTools.log(log,
					new BetriebsmeldungDaten(anfrage.getMessQuerschnitt().getSystemObject()),
					GlProgMsg.WarningKalenderAntwortetNicht, anfrage.getMessQuerschnitt());
			kalenderantwort = Collections.emptyList();
		}

		return kalenderantwort;
	}

}
